﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using System.IO;
using System.Drawing;
using System.Drawing.Imaging;

using System.Runtime.InteropServices;

// See: How to create an Icon file that contains Multiple Sizes / Images in C#
//      https://stackoverflow.com/questions/3213999/how-to-create-an-icon-file-that-contains-multiple-sizes-images-in-c-sharp

namespace howto_make_icon_from_files
{
    public class IconFactory
    {
        [DllImport("shell32.dll")]
        static extern void SHChangeNotify(
            HChangeNotifyEventID wEventId,
            HChangeNotifyFlags uFlags,
            IntPtr dwItem1,
            IntPtr dwItem2);
        enum HChangeNotifyFlags : uint
        {
            SHCNF_IDLIST = 0x0000,
            SHCNF_PATHA = 0x0001,
            SHCNF_PRINTERA = 0x0002,
            SHCNF_DWORD = 0x0003,
            SHCNF_PATHW = 0x0005,
            SHCNF_PRINTERW = 0x0006,
            SHCNF_TYPE = 0x00FF,
            SHCNF_FLUSH = 0x1000,
            SHCNF_FLUSHNOWAIT = 0x2000,
            SHCNF_PATH = SHCNF_PATHW,
            SHCNF_PRINTER = SHCNF_PRINTERW
        }

        [Flags]
        enum HChangeNotifyEventID
        {
            SHCNE_ALLEVENTS = 0x7FFFFFFF,
            SHCNE_ASSOCCHANGED = 0x08000000,
            SHCNE_ATTRIBUTES = 0x00000800,
            SHCNE_CREATE = 0x00000002,
            SHCNE_DELETE = 0x00000004,
            SHCNE_DRIVEADD = 0x00000100,
            SHCNE_DRIVEADDGUI = 0x00010000,
            SHCNE_DRIVEREMOVED = 0x00000080,
            SHCNE_EXTENDED_EVENT = 0x04000000,
            SHCNE_FREESPACE = 0x00040000,
            SHCNE_MEDIAINSERTED = 0x00000020,
            SHCNE_MEDIAREMOVED = 0x00000040,
            SHCNE_MKDIR = 0x00000008,
            SHCNE_NETSHARE = 0x00000200,
            SHCNE_NETUNSHARE = 0x00000400,
            SHCNE_RENAMEFOLDER = 0x00020000,
            SHCNE_RENAMEITEM = 0x00000001,
            SHCNE_RMDIR = 0x00000010,
            SHCNE_SERVERDISCONNECT = 0x00004000,
            SHCNE_UPDATEDIR = 0x00001000,
            SHCNE_UPDATEIMAGE = 0x00008000,
        }

        public const int MaxIconWidth = 256;
        public const int MaxIconHeight = 256;
        private const ushort HeaderReserved = 0;
        private const ushort HeaderIconType = 1;
        private const byte HeaderLength = 6;
        private const byte EntryReserved = 0;
        private const byte EntryLength = 16;
        private const byte PngColorsInPalette = 0;
        private const ushort PngColorPlanes = 1;

        // Return an icon as an Icon object.
        public static Icon SavePngsAsIcon(
            IEnumerable<Bitmap> images)
        {
            using (MemoryStream stream = new MemoryStream())
            {
                // Write the icon into the stream.
                SavePngsAsIcon(images, stream, false);

                // Create a new stream from the first one.
                using (MemoryStream stream2 =
                    new MemoryStream(stream.ToArray()))
                {

                    // Create and return an icon from the stream.
                    return new Icon(stream2);
                }
            }
        }

        // Save an icon into a file.
        public static void SavePngsAsIcon(
            IEnumerable<Bitmap> images,
            string filename, bool notify_system)
        {
            using (FileStream stream = new FileStream(filename, FileMode.Create))
            {
                SavePngsAsIcon(images, stream, notify_system);
            }

        }

        // Write an icon into a stream.
        // Note that this closes the stream.
        public static void SavePngsAsIcon(
            IEnumerable<Bitmap> images,
            Stream stream, bool notify_system)
        {
            Bitmap[] ordered_images =
                images.OrderBy(i => i.Width).ThenBy(i => i.Height).ToArray();
            using (BinaryWriter writer = new BinaryWriter(stream))
            {
                writer.Write(HeaderReserved);
                writer.Write(HeaderIconType);
                writer.Write((ushort)ordered_images.Length);

                Dictionary<uint, byte[]> buffers =
                    new Dictionary<uint, byte[]>();
                uint length_sum = 0;
                uint base_offset = (uint)(HeaderLength * EntryLength * ordered_images.Length);
                for (int i = 0; i < ordered_images.Length; i++)
                {
                    Bitmap image = ordered_images[i];
                    byte[] buffer = CreateImageBuffer(image);
                    uint offset = base_offset + length_sum;

                    writer.Write(GetIconWidth(image));
                    writer.Write(GetIconHeight(image));
                    writer.Write(PngColorsInPalette);
                    writer.Write(EntryReserved);
                    writer.Write(PngColorPlanes);
                    writer.Write((ushort)Image.GetPixelFormatSize(image.PixelFormat));
                    writer.Write((uint)buffer.Length);
                    writer.Write(offset);

                    length_sum += (uint)buffer.Length;
                    buffers.Add(offset, buffer);
                }

                foreach (KeyValuePair<uint, byte[]> kvp in buffers)
                {
                    writer.BaseStream.Seek(kvp.Key, SeekOrigin.Begin);
                    writer.Write(kvp.Value);
                }
            }

            if (notify_system)
            {
                SHChangeNotify(
                    HChangeNotifyEventID.SHCNE_ASSOCCHANGED,
                    HChangeNotifyFlags.SHCNF_IDLIST,
                    IntPtr.Zero, IntPtr.Zero);
            }
        }

        private static byte GetIconWidth(Bitmap image)
        {
            if (image.Width >= MaxIconWidth) return 0;
            return (byte)image.Width;
        }

        private static byte GetIconHeight(Bitmap image)
        {
            if (image.Height >= MaxIconHeight) return 0;
            return (byte)image.Height;
        }

        private static byte[] CreateImageBuffer(Bitmap image)
        {
            using (MemoryStream stream = new MemoryStream())
            {
                image.Save(stream, ImageFormat.Png);
                return stream.ToArray();
            }
        }
    }
}
